@@ -3,7 +3,6 @@
 #define _PYGOBJECT_H_
 
 #include <Python.h>
-#include <ExtensionClass.h>
 
 #include <glib.h>
 #include <glib-object.h>
@@ -13,10 +12,11 @@ typedef struct {
     GObject *obj;
     gboolean hasref;     /* the GObject owns this reference */
     PyObject *inst_dict; /* the instance dictionary -- must be last */
+    PyObject *weakreflist; /* list of weak references */
 } PyGObject;
 
 #define pygobject_get(v) (((PyGObject *)(v))->obj)
-#define pygobject_check(v,base) (ExtensionClassSubclassInstance_Check(v,base))
+#define pygobject_check(v,base) (PyObject_TypeCheck(v,base))
 
 typedef struct {
     PyObject_HEAD
@@ -26,13 +26,13 @@ typedef struct {
 } PyGBoxed;
 
 #define pyg_boxed_get(v,t)      ((t *)((PyGBoxed *)(v))->boxed)
-#define pyg_boxed_check(v,typecode) (ExtensionClassSubclassInstance_Check(v, &PyGBoxed_Type) && ((PyGBoxed *)(v))->gtype == typecode)
+#define pyg_boxed_check(v,typecode) (PyObject_TypeCheck(v, &PyGBoxed_Type) && ((PyGBoxed *)(v))->gtype == typecode)
 
 struct _PyGObject_Functions {
     void (* register_class)(PyObject *dict, const gchar *class_name,
-			    GType type, PyExtensionClass *ec, PyObject *bases);
+			    GType gtype, PyTypeObject *type, PyObject *bases);
     void (* register_wrapper)(PyObject *self);
-    PyExtensionClass *(* lookup_class)(GType type);
+    PyTypeObject *(* lookup_class)(GType type);
     PyObject *(* newgobj)(GObject *obj);
     GClosure *(* closure_new)(PyObject *callback, PyObject *extra_args,
 			      PyObject *swap_data);
@@ -49,11 +49,11 @@ struct _PyGObject_Functions {
     PyObject *(* value_as_pyobject)(const GValue *value);
 
     void (* register_interface)(PyObject *dict, const gchar *class_name,
-				GType type, PyExtensionClass *ec);
+				GType gtype, PyTypeObject *type);
 
-    PyExtensionClass *boxed_type;
+    PyTypeObject *boxed_type;
     void (* register_boxed)(PyObject *dict, const gchar *class_name,
-			    GType boxed_type, PyExtensionClass *ec);
+			    GType boxed_type, PyTypeObject *type);
     PyObject *(* boxed_new)(GType boxed_type, gpointer boxed,
 			    gboolean copy_boxed, gboolean own_ref);
 
@@ -96,7 +96,7 @@ struct _PyGObject_Functions *_PyGObject_API;
         PyObject *mdict = PyModule_GetDict(gobject); \
         PyObject *cobject = PyDict_GetItemString(mdict, "_PyGObject_API"); \
         if (PyCObject_Check(cobject)) \
-            _PyGObject_API = PyCObject_AsVoidPtr(cobject); \
+            _PyGObject_API = (struct _PyGObject_Functions *)PyCObject_AsVoidPtr(cobject); \
         else { \
 	    Py_FatalError("could not find _PyGObject_API object"); \
 	    return; \
@@ -105,7 +105,6 @@ struct _PyGObject_Functions *_PyGObject_API;
         Py_FatalError("could not import gobject"); \
         return; \
     } \
-    ExtensionClassImported; \
 }
 
 #endif /* !_INSIDE_PYGOBJECT_ */
